# User interface
inputDataUploadUI <- function(id)
{
  ns <- NS(id)

  tagList(
    div(style = "margin-top: 15px"),
    box(
      width = 12,
      title = "Input data",
      solidHeader = FALSE,
      status = "primary",
      collapsible = TRUE,
      fluidRow(
        column(width = 4,
               fileInput(ns("fileInput"),
                         width = "100%",
                         label = "File input:"),
               p("Maximum file size: 70MB",
                 tags$br(),
                 "Supported files types: txt, csv, xls, xlsx (uncompressed and zip archives)")),
        column(width = 8,
               withSpinner(uiOutput(ns("origDataDetails")),
                           type = 7,
                           proxy.height = "50px")))),
    uiOutput(ns("attrMapping"))
  )
}

# Server logic
inputDataUpload <- function(input, output, session)
{
  # Get namespace
  ns <- session$ns

  vals <- reactiveValues(attrMapping = list(),
                         attrMappingStatus = list(),
                         attrMappingValid = FALSE,
                         origDataAttrs = c(),
                         inputData = NULL)

  # EVENT: Input data file name changed
  originalData <- eventReactive(input[["fileInput"]], {
    # Get reference to file input
    fileInput <- input$fileInput

    # Validate
    validate(need(fileInput, message = FALSE))

    # Read input data
    originalData <- ReadDataFile(fileInput$datapath)

    vals$origDataAttrs <- colnames(originalData)
    vals$attrMapping <- GetPreliminaryAttributesMapping(originalData)

    if ("FirstCD4Count" %in% names(vals$attrMapping) &
        "cd4_num" %in% colnames(originalData)) {
      vals$attrMapping[["FirstCD4Count"]] <- "cd4_num"
    }

    return(originalData)
  }, ignoreNULL = TRUE)

  # Output original data details when they have changed
  output[["origDataDetails"]] <- renderUI({
    # Respond to originalData change
    originalData <- originalData()

    isolate({
      # Get reference to file input
      fileInput <- input$fileInput

      # Get input data details
      inputDataDetails <-
        fluidRow(
          column(3,
                 p(strong("Name:"), fileInput$name),
                 p(strong("Size:"), FormatObjectSize(fileInput$size)),
                 p(strong("Type:"), fileInput$type),
                 p(strong("Number of records:"), nrow(originalData))),
          column(9,
                 p(strong("Column names:")),
                 p(id = ns("colNamesP"),
                   paste(colnames(originalData), collapse = ", ")))
        )

      return(inputDataDetails)
    })
  })

  # Generate attributes mapping
  output[["attrMapping"]] <- renderUI({
    # Respond to originalData change
    originalData()

    isolate({

      ns <- session$ns

      attrMappingStatusUI <- uiOutput(ns("attrMappingStatus"))
      valueCheckStatusUI <- uiOutput(ns("valueCheckStatus"))
      descrText <- p("Please, provide mapping between attributes used internally by the tool (column \"Tool\") and data dimensions (column \"Data\").")

      attributesMappingRowsUI <- lapply(names(vals$attrMapping), function(analysisAttr) {
        selectOptions <- list(placeholder = paste("Select", analysisAttr, "attribute"))
        if (is.null(vals$attrMapping[[analysisAttr]])) {
          selectOptions <- list(
            placeholder = paste("Select", analysisAttr, "attribute"),
            onInitialize = I('function() { this.setValue(""); }'))
          validMsg <- "Not ok"
        }

        selectId <- paste0("attr_", analysisAttr)

        widget <- tags$tr(
          # Selection attribute
          tags$td(analysisAttr),
          # Selection data dimension
          tags$td(selectizeInput(ns(selectId),
                                 label = NULL,
                                 choices = vals$origDataAttrs,
                                 selected = vals$attrMapping[[analysisAttr]],
                                 options = selectOptions))
        )

        # EVENT: Selected item changed
        observeEvent(input[[selectId]], {
          vals$attrMapping[[analysisAttr]] <- input[[selectId]]
        })

        return(widget)
      })

      attributesMappingTableUI <- tags$table(
        id = ns("attrMappingTable"),
        tags$thead(
          tags$tr(
            tags$th(class = ns("attrNameCol"),
                    "Tool"),
            tags$th("Data")
          )
        ),
        tags$tbody(
          attributesMappingRowsUI
        )
      )

      # Wrap in a box
      attributesMappingUI <- box(
        width = 12,
        title = "Attributes mapping",
        solidHeader = FALSE,
        status = "warning",
        collapsible = TRUE,
        descrText,
        fluidRow(
          column(6, attributesMappingTableUI),
          column(6, attrMappingStatusUI, valueCheckStatusUI)
        ))

      return(attributesMappingUI)
    })
  })

  observeEvent(vals$attrMapping, {
    attrMappingStatus <- GetAttrMappingStatus(vals$attrMapping)
    vals$attrMappingValid <- attrMappingStatus$Valid
    vals$attrMappingStatus <- attrMappingStatus[c("NonMapped", "MultipleMapped")]
  })

  output[["attrMappingStatus"]] <- renderUI({
    nonMappedAnalysisAttrs <- vals$attrMappingStatus$NonMapped
    multipleMappedDataAttrs <- vals$attrMappingStatus$MultipleMapped
    attrMappingValid <- vals$attrMappingValid

    isolate({

      header <- NULL
      textColor <- NULL
      if (attrMappingValid) {
        textColor <- "seagreen"
        headerHTML <- p(tags$i(class = "material-icons", "check"), "Mapping is valid.", style = paste("color:", textColor))
      } else {
        textColor <- "red"
        headerHTML <- p(tags$i(class = "material-icons", "error_outline"), "Mapping is not valid. Please, correct the following issues:", style = paste("color:", textColor))
      }

      nonMappedStatusHTML <- NULL
      if (length(nonMappedAnalysisAttrs) > 0) {
        nonMappedStatusHTML <- tagList(
          tags$span("Tool attributes not mapped:", style = "font-weight: bold"),
          tags$ol(
            lapply(nonMappedAnalysisAttrs, function(analysisAttr) tags$li(analysisAttr))
          )
        )
      }

      multiMappedStatusHTML <- NULL
      if (length(multipleMappedDataAttrs) > 0) {
        multiMappedStatusHTML <- tagList(
          tags$span("Data attributes mapped multiple times:", style = "font-weight: bold"),
          tags$ol(
            lapply(names(multipleMappedDataAttrs), function(dataAttr) {
              analysisAttrs <- multipleMappedDataAttrs[[dataAttr]]
              tags$li(paste(AddQuoteMarks(dataAttr),
                            "mapped to",
                            paste(AddQuoteMarks(analysisAttrs), collapse = ", ")))
            })
          )
        )
      }

      attrMappingStatusHTML <- box(width = 12,
                                   style = "background-color: ghostwhite;",
                                   headerHTML,
                                   nonMappedStatusHTML,
                                   multiMappedStatusHTML)

      return(attrMappingStatusHTML)
    })
  })

  inputData <- eventReactive(vals$attrMapping, {
    intermediateData <- ApplyAttributesMapping(originalData(), vals$attrMapping)
    return(PreProcessInputData(intermediateData))
  })

  output[["valueCheckStatus"]] <- renderUI({
    # Respond to inputData change
    inputData <- inputData()

    isolate({
      if (is.null(inputData)) {
        return(NULL)
      }
      inputDataValidityInfo <- ValidateInputData(inputData)

      inputDataValid <- all(sapply(inputDataValidityInfo, "[[", "Valid"))
      wrongValuesCols <- Filter(function(x) length(x$WrongValues) != 0, inputDataValidityInfo)

      headerHTML <- NULL
      textColor <- NULL
      if (inputDataValid) {
        textColor <- "seagreen"
        headerHTML <- p(tags$i(class = "material-icons", "check"),
                        "Values are valid.",
                        style = paste("color:", textColor))
      } else {
        textColor <- "red"
        headerHTML <- p(tags$i(class = "material-icons", "error_outline"),
                        "Values are not valid. Please, correct the following issues:",
                        style = paste("color:", textColor))
      }

      wrongValuesHTML <- NULL
      if (length(wrongValuesCols) > 0) {
        wrongValuesHTML <- tags$li(
          "Incorrect values:",
          tags$ul(
            lapply(names(wrongValuesCols), function(colName) {
              wrongValuesCol <- wrongValuesCols[[colName]]
              tags$li(paste(AddQuoteMarks(colName),
                            "contains invalid value(s)",
                            paste(AddQuoteMarks(wrongValuesCol$WrongValues), collapse = ", ")))
            })
          )
        )
      }

      valueCheckStatusHTML <- box(width = 12,
                                  style = "background-color: ghostwhite;",
                                  headerHTML,
                                  tags$ol(wrongValuesHTML,
                                          style = paste("color:", textColor)))

      return(valueCheckStatusHTML)
    })
  })

  return(list(originalData = originalData,
              inputData = inputData))
}
